==> luna_cow-0.7.8.tar.gz: luna_cow-0.7.8/cow/redis_db/redis_listener_context.py <==
"""
Module with redis listener contexts
"""

import asyncio
from contextlib import asynccontextmanager
from time import time
from typing import ContextManager, Type

import redis
from async_timeout import timeout as async_timeout
from redis.asyncio import Redis

from ..errors.errors import Error
from ..monitoring.manager import LunaMonitoringManager
from ..redis_db.redis_entities import BaseWSConnection, ChannelPublisher, RedisChannel, RedisDisconnected
from ..utils.log import Logger
from .adaptor import RedisAdaptor
from .monitoring_points import ErrorPoint


class RedisContext:
    """
    Base class for RedisContext

    Holds logic for auto-reconnecting redis consumer.

    Attributes:
        subscriptions (Dict[str, ChannelPublisher]): Mapping between channel name and active publishers
        channelName (str): channel name for subscription
        logger (Logger): logger for subscription
        resurrectingRedisTask: task for monitoring messages in redis channel
        redis: redis-py command interface
        resurrectingRedisTask: task for maintaining redis connection
        timeout: timeout for single subscribe operations (actually, we do multiple attempts)
        pingInterval: ping interval
        monitoring: monitoring client
        subscribeAttempts: subscribe attempts
        reconnectionDelay: reconnection delay
        _channelPublisher: channel publisher class
        _locks: asyncio locks whose prevent simanteniously new subscription addition and old subscription deletion
                for each channel
        adaptor: redis adaptor, container for redis client
    """

    def __init__(  # pylint: disable-msg=too-many-arguments
        self,
        *,
        host: str = None,
        port: int = None,
        username: str = None,
        password: str = None,
        dbNumber: int = None,
        sentinels: list[tuple[str, int]] = None,
        sentinelUsername: str = None,
        sentinelPassword: str = None,
        masterName: str = None,
        channelName: str,
        monitoring: LunaMonitoringManager,
        commandTimeout: int,
        pingInterval: int,
        subscribeAttempts: int,
        reconnectionDelay: int,
        channelPublisher: Type[ChannelPublisher],
    ) -> None:  # pylint: disable-msg=too-many-arguments
        """Initialize context"""
        self.subscriptions: dict[str, ChannelPublisher] = {}  # Map chanel name to chanel publishers
        self.channelName = channelName
        self.logger = Logger("redis_db")

        self.resurrectingRedisTask = asyncio.create_task(self.resurrectingRedis())
        self.timeout = commandTimeout
        self.pingInterval = pingInterval
        self.monitoring = monitoring
        self.subscribeAttempts = subscribeAttempts
        self.reconnectionDelay = reconnectionDelay
        self._channelPublisher = channelPublisher
        self._locks: dict[str, asyncio.Lock] = {}
        self.adaptor = RedisAdaptor(  # pylint: disable-msg=duplicate-code
            host=host,
            port=port,
            username=username,
            password=password,
            sentinels=sentinels,
            sentinelUsername=sentinelUsername,
            sentinelPassword=sentinelPassword,
            masterName=masterName,
            dbNumber=dbNumber,
        )

    @property
    def redis(self) -> Redis:
        """Get redis client"""
        return self.adaptor.client

    async def resurrectingRedis(self) -> None:
        """Auto-reconnecting consumer for redis."""
        while True:
            self.logger.info("init connection to redis")
            try:
                # Connect to redis. If we had subscriptions before - resubscribe.
                channels = [publisher.redisChannel for publisher in self.subscriptions.values()]
                self.logger.info(f"Subscribing to {channels}")
                for publisher in self.subscriptions.values():
                    async with async_timeout(self.timeout):
                        await publisher.reconnect()

                # Periodically send PING commands to ensure that redis is alive
                while True:
                    async with async_timeout(self.timeout):
                        await self.redis.ping()
                    await asyncio.sleep(self.pingInterval)

            except asyncio.TimeoutError:
                self.logger.error("Timeout occured, reconnecting...")
            except redis.exceptions.ConnectionError:
                self.logger.error("Connection closed, reconnecting...")
            except redis.exceptions.RedisError as e:
                self.logger.error(f"Redis error occured: {e}")
            except ConnectionError as e:
                self.logger.error(f"Failed to connect to redis: {e}")
            except asyncio.CancelledError:
                self.logger.debug("Cancelling redis connecion")
                break
            except Exception:  # pylint: disable-msg=broad-except
                self.logger.exception()

            point = ErrorPoint(time(), errorCode=Error.RedisDisconnected.errorCode)
            self.monitoring.flushPoints([point])

            # Prevent flooding
            await asyncio.sleep(self.reconnectionDelay)

        if self.adaptor:
            await self.adaptor.close()
            self.adaptor = None

    async def close(self) -> None:
        """Close all publishers and redis"""
        tasks = [subscription.closePublisher(self.timeout) for subscription in self.subscriptions.values()]
        await asyncio.gather(*tasks)

        if self.resurrectingRedisTask is not None:
            self.resurrectingRedisTask.cancel()

            # Gotta await, because we handle CancelledError inside the task - to close redis connections
            await self.resurrectingRedisTask
            self.resurrectingRedisTask = None

    @asynccontextmanager
    async def subscribe(self, wsConnection: BaseWSConnection, accountId: str, requestLogger: Logger) -> ContextManager:
        """
        Context Manager, subscription to events from luna service.

        Args:
            wsConnection: ws connection
            accountId: account id
            requestLogger: logger with appropriate requestId

        Raises:
            RedisDisconnected: If subscribe failed for some reason
        """
        ch = f"{self.channelName}:{accountId}"
        try:
            await self.addSubscription(ch, wsConnection)
            yield
        except RedisDisconnected:
            error = Error.FailToSubscribe.format(ch)
            requestLogger.exception(f"fail to subscribe to a channel, {error.asDict()}")
            self.monitoring.flushPoints([ErrorPoint(time(), errorCode=error.errorCode)])
            raise
        finally:
            await asyncio.shield(self.removeSubscription(ch, wsConnection))

    async def addSubscription(self, ch: str, wsConnection: BaseWSConnection) -> None:
        """
        Add connection to channel subscribers list.

        If no channel exists creates new.

        Args:
            ch: A str with channel name
            wsConnection: A WSConnection instance

        Raises:
            RedisDisconnected: If subscribe failed for some reason
        """
        self.logger.debug(f"Add subscription to channel {ch}, {wsConnection.subscriptionId}")

        if ch not in self._locks:
            self._locks[ch] = asyncio.Lock()

        try:
            async with self._locks[ch]:
                if ch not in self.subscriptions:
                    redisChannel = RedisChannel(ch, self.redis.pubsub())
                    publisher = self._channelPublisher(ch, redisChannel, monitoring=self.monitoring)
                    await publisher.subscribeToRedisChannel(self.subscribeAttempts, self.timeout)
                    self.subscriptions[ch] = publisher
                self.subscriptions[ch][wsConnection.subscriptionId] = wsConnection
        finally:
            if ch in self._locks:
                del self._locks[ch]

    async def removeSubscription(self, ch: str, wsConnection: BaseWSConnection) -> None:
        """
        Remove connection from channel subscribers list.

        Closes channel if no more subscribers left.

        Args:
            ch: A str with channel name
            wsConnection: A WSConnection instance
        """
        self.logger.debug(f"Remove subscription to channel {ch}, {wsConnection.subscriptionId}")
        if ch not in self.subscriptions:
            return
        if wsConnection.subscriptionId not in self.subscriptions[ch]:
            return

        if ch not in self._locks:
            self._locks[ch] = asyncio.Lock()

        try:
            async with self._locks[ch]:
                del self.subscriptions[ch][wsConnection.subscriptionId]
                publisher = self.subscriptions[ch]
                if not publisher.hasSubsribers:
                    await publisher.closePublisher(self.timeout)
                    del self.subscriptions[ch]

        finally:
            if ch in self._locks:
                del self._locks[ch]

==> luna_cow-0.8.30.tar.gz: luna_cow-0.8.30/cow/redis_db/redis_listener_context.py <==
"""
Module with redis listener contexts
"""

import asyncio
from asyncio import timeout as async_timeout
from contextlib import asynccontextmanager
from time import time
from typing import ContextManager, Type

import redis
from redis.asyncio import Redis

from ..errors.errors import Error
from ..monitoring.manager import LunaMonitoringManager
from ..redis_db.redis_entities import BaseWSConnection, ChannelPublisher, RedisChannel, RedisDisconnected
from ..utils.log import Logger
from .adaptor import RedisAdaptor
from .monitoring_points import ErrorPoint


class RedisContext:
    """
    Base class for RedisContext

    Holds logic for auto-reconnecting redis consumer.

    Attributes:
        subscriptions (Dict[str, ChannelPublisher]): Mapping between channel name and active publishers
        channelName (str): channel name for subscription
        logger (Logger): logger for subscription
        resurrectingRedisTask: task for monitoring messages in redis channel
        redis: redis-py command interface
        resurrectingRedisTask: task for maintaining redis connection
        timeout: timeout for single subscribe operations (actually, we do multiple attempts)
        pingInterval: ping interval
        monitoring: monitoring client
        subscribeAttempts: subscribe attempts
        reconnectionDelay: reconnection delay
        _channelPublisher: channel publisher class
        _locks: asyncio locks whose prevent simanteniously new subscription addition and old subscription deletion
                for each channel
        adaptor: redis adaptor, container for redis client
    """

    def __init__(  # pylint: disable-msg=too-many-arguments
        self,
        *,
        host: str = None,
        port: int = None,
        username: str = None,
        password: str = None,
        dbNumber: int = None,
        sentinels: list[tuple[str, int]] = None,
        sentinelUsername: str = None,
        sentinelPassword: str = None,
        masterName: str = None,
        channelName: str,
        monitoring: LunaMonitoringManager,
        commandTimeout: int,
        pingInterval: int,
        subscribeAttempts: int,
        reconnectionDelay: int,
        channelPublisher: Type[ChannelPublisher],
    ) -> None:  # pylint: disable-msg=too-many-arguments
        """Initialize context"""
        self.subscriptions: dict[str, ChannelPublisher] = {}  # Map chanel name to chanel publishers
        self.channelName = channelName
        self.logger = Logger("redis_db")

        self.resurrectingRedisTask = asyncio.create_task(self.resurrectingRedis())
        self.timeout = commandTimeout
        self.pingInterval = pingInterval
        self.monitoring = monitoring
        self.subscribeAttempts = subscribeAttempts
        self.reconnectionDelay = reconnectionDelay
        self._channelPublisher = channelPublisher
        self._locks: dict[str, asyncio.Lock] = {}
        self.adaptor = RedisAdaptor(  # pylint: disable-msg=duplicate-code
            host=host,
            port=port,
            username=username,
            password=password,
            sentinels=sentinels,
            sentinelUsername=sentinelUsername,
            sentinelPassword=sentinelPassword,
            masterName=masterName,
            dbNumber=dbNumber,
        )

    @property
    def redis(self) -> Redis:
        """Get redis client"""
        return self.adaptor.client

    async def resurrectingRedis(self) -> None:
        """Auto-reconnecting consumer for redis."""
        while True:
            self.logger.info("init connection to redis")
            try:
                # Connect to redis. If we had subscriptions before - resubscribe.
                channels = [publisher.redisChannel for publisher in self.subscriptions.values()]
                self.logger.info(f"Subscribing to {channels}")
                for publisher in self.subscriptions.values():
                    async with async_timeout(self.timeout):
                        await publisher.reconnect()

                # Periodically send PING commands to ensure that redis is alive
                while True:
                    async with async_timeout(self.timeout):
                        await self.redis.ping()
                    await asyncio.sleep(self.pingInterval)

            except asyncio.TimeoutError:
                self.logger.error("Timeout occurred, reconnecting...")
            except redis.exceptions.ConnectionError:
                self.logger.error("Connection closed, reconnecting...")
            except redis.exceptions.RedisError as e:
                self.logger.error(f"Redis error occurred: {e}")
            except ConnectionError as e:
                self.logger.error(f"Failed to connect to redis: {e}")
            except asyncio.CancelledError:
                self.logger.debug("Cancelling redis connection")
                break
            except Exception:  # pylint: disable-msg=broad-except
                self.logger.exception()

            point = ErrorPoint(time(), errorCode=Error.RedisDisconnected.errorCode)
            self.monitoring.flushPoints([point])

            # Prevent flooding

            try:
                await asyncio.sleep(self.reconnectionDelay)
            except asyncio.CancelledError:
                self.logger.debug("Cancelling redis connection")
                break

        if self.adaptor:
            await self.adaptor.close()
            self.adaptor = None

    async def close(self) -> None:
        """Close all publishers and redis"""
        tasks = [subscription.closePublisher(self.timeout) for subscription in self.subscriptions.values()]
        await asyncio.gather(*tasks)

        if self.resurrectingRedisTask is not None:
            self.resurrectingRedisTask.cancel()

            # Gotta await, because we handle CancelledError inside the task - to close redis connections
            await self.resurrectingRedisTask
            self.resurrectingRedisTask = None

    @asynccontextmanager
    async def subscribe(self, wsConnection: BaseWSConnection, accountId: str, requestLogger: Logger) -> ContextManager:
        """
        Context Manager, subscription to events from luna service.

        Args:
            wsConnection: ws connection
            accountId: account id
            requestLogger: logger with appropriate requestId

        Raises:
            RedisDisconnected: If subscribe failed for some reason
        """
        ch = f"{self.channelName}:{accountId}"
        try:
            await self.addSubscription(ch, wsConnection)
            yield
        except RedisDisconnected:
            error = Error.FailToSubscribe.format(ch)
            requestLogger.exception(f"fail to subscribe to a channel, {error.asDict()}")
            self.monitoring.flushPoints([ErrorPoint(time(), errorCode=error.errorCode)])
            raise
        finally:
            await asyncio.shield(self.removeSubscription(ch, wsConnection))

    async def addSubscription(self, ch: str, wsConnection: BaseWSConnection) -> None:
        """
        Add connection to channel subscribers list.

        If no channel exists creates new.

        Args:
            ch: A str with channel name
            wsConnection: A WSConnection instance

        Raises:
            RedisDisconnected: If subscribe failed for some reason
        """
        self.logger.debug(f"Add subscription to channel {ch}, {wsConnection.subscriptionId}")

        if ch not in self._locks:
            self._locks[ch] = asyncio.Lock()

        try:
            async with self._locks[ch]:
                if ch not in self.subscriptions:
                    redisChannel = RedisChannel(ch, self.redis.pubsub())
                    publisher = self._channelPublisher(ch, redisChannel, monitoring=self.monitoring)
                    await publisher.subscribeToRedisChannel(self.subscribeAttempts, self.timeout)
                    self.subscriptions[ch] = publisher
                self.subscriptions[ch][wsConnection.subscriptionId] = wsConnection
        finally:
            if ch in self._locks:
                del self._locks[ch]

    async def removeSubscription(self, ch: str, wsConnection: BaseWSConnection) -> None:
        """
        Remove connection from channel subscribers list.

        Closes channel if no more subscribers left.

        Args:
            ch: A str with channel name
            wsConnection: A WSConnection instance
        """
        self.logger.debug(f"Remove subscription to channel {ch}, {wsConnection.subscriptionId}")
        if ch not in self.subscriptions:
            return
        if wsConnection.subscriptionId not in self.subscriptions[ch]:
            return

        if ch not in self._locks:
            self._locks[ch] = asyncio.Lock()

        try:
            async with self._locks[ch]:
                del self.subscriptions[ch][wsConnection.subscriptionId]
                publisher = self.subscriptions[ch]
                if not publisher.hasSubsribers:
                    await publisher.closePublisher(self.timeout)
                    del self.subscriptions[ch]

        finally:
            if ch in self._locks:
                del self._locks[ch]
